#include "config.h"

/******************************************/
#ifdef VON_NEUMANN
#define NEIGHBORS    4

#define TOP          0
#define RIGHT        1
#define BOTTOM       2
#define LEFT         3
#endif

#ifdef MOORE
#define NEIGHBORS    8

#define TOP          0
#define TOP_RIGHT    1
#define RIGHT        2
#define BOTTOM_RIGHT 3
#define BOTTOM       4
#define BOTTOM_LEFT  5
#define LEFT         6
#define TOP_LEFT     7
#endif

// special status 
#define QUIESCENT    8
#define WALL         9

// no information
#define NONE        -1

/******************************************/
typedef struct info_t
{
   int      origin;
   int      status;
   int      state;
   int      fired;
   int      new_state;
   int      new_status;
}
*info_t;

info_t   make_info (int origin,int status,int state);

/******************************************/
typedef info_t neighborhood_t[NEIGHBORS];

void clear_neighborhood (neighborhood_t);
void set_neighbor (neighborhood_t,info_t);

/********************************************************/
#define OPPOSITE(dir) (((dir)+NEIGHBORS/2)%NEIGHBORS)
#define INVCLK(dir)   (((dir)+1)%NEIGHBORS)
#define CLK(dir)      (((dir)+3)%NEIGHBORS)

/********************************************************/
// Self-Replicating loops
/********************************************************/
#define BASIC        0
#define COLLIDE      1
#define BARRIER      2
#define GROW_GENE    3
#define CLK_GENE     4
#define INVCLK_GENE  5
#define TRANSLATOR   6
#define SPROUT       7
#define WAITER       8
#define PRE_WAITER   9
#define STOP        10
#define ERASE       11

#define STATE_OF_NEIGHBOR(dir) (d->neighborhood[dir]->state)
#define STATUS_OF_NEIGHBOR(dir) (d->neighborhood[dir]->status)

#define NEIGHBOR_STATE(dir,v) \
  (d->neighborhood[dir] && STATE_OF_NEIGHBOR(dir) == v)

#define NEIGHBOR_STATUS(dir,v) \
  (d->neighborhood[dir] && STATUS_OF_NEIGHBOR(dir) == v)

#define DEAD_NEIGHBOR(dir) \
  (d->neighborhood[dir] == NULL ||\
   STATUS_OF_NEIGHBOR(dir) == QUIESCENT ||\
   STATUS_OF_NEIGHBOR(dir) == WALL ||\
   STATE_OF_NEIGHBOR(dir) == ERASE ||\
   STATE_OF_NEIGHBOR(dir) == STOP \
)

#define TURNING_GENE_IN(dir)\
   (NEIGHBOR_STATE(dir,INVCLK_GENE) || NEIGHBOR_STATE(dir,CLK_GENE))

#define IS_TURN_GENE(v) (v==CLK_GENE || v==INVCLK_GENE)

/********************************************************/
int cell_number;

/********************************************************/
int context [NEIGHBORS][2];

int inside_applet (int x,int y);
int color_of (int state);

void cellular_automaton_init ();

/********************************************************/
#define NORMALIZE(x,y) \
   if (x >= MAXX) x -= MAXX; \
   else if (x < 0) x += MAXX;\
   if (y >= MAXY) y -= MAXY; \
   else if (y < 0) y += MAXY;

/********************************************************/
